#include <Python.h>

PyObject *pyace(PyObject *self, PyObject* args);
PyObject *pyblackfilter(PyObject *self, PyObject* args);
PyObject *pyblurfilter(PyObject *self, PyObject* args);
PyObject *pyborder(PyObject *self, PyObject* args);
PyObject *pycanny(PyObject *self, PyObject* args);
PyObject *pycompare(PyObject *self, PyObject* args);
PyObject *pygaussian(PyObject *self, PyObject* args);
PyObject *pygrayfilter(PyObject *self, PyObject* args);
PyObject *pymasks(PyObject *self, PyObject* args);
PyObject *pynoisefilter(PyObject *self, PyObject* args);
PyObject *pyfind_scan_borders(PyObject *self, PyObject* args);
PyObject *pysobel(PyObject *self, PyObject* args);
PyObject *pyswt(PyObject *self, PyObject* args);
PyObject *pyget_version(PyObject *self, PyObject* args);
